/***
*clutil.cxx - Class Lib component-wide utility functions.
*
*  Copyright (C) 1991, Microsoft Corporation.  All Rights Reserved.
*  Information Contained Herein Is Proprietary and Confidential.
*
*Purpose:
*  typelib utility functions
*
*Revision History:
*   [00] 24-Jan-93 RajivK:	Created
*
*Implementation Notes:
*
*****************************************************************************/

#include "typelib.hxx"
#include "silver.hxx"
#include <xstring.h>
#include <time.h>
#include <ctype.h>	    // for isspace() et al.


#pragma hdrstop(RTPCHNAME)

#if ID_DEBUG
#undef SZ_FILE_NAME
#if OE_MAC
static char szOleTlibutilCxx[] = __FILE__;
#define SZ_FILE_NAME szOleTlibutilCxx
#else 
static char szTlibutilCxx[] = __FILE__;
#define SZ_FILE_NAME szTlibutilCxx
#endif 
#endif 


// forward declarations for DBCS processing

ACHAR MapDBChar(ACHAR xch, LCID lcid);
unsigned long LHashValOfDBCSName(SYSKIND syskind, LCID lcid, const char FAR* szName);


// Define an array for mapping all the characters.
//
#pragma code_seg(CS_CP1252)
ALLOC_CODE(BYTE) g_rgbCodePage1252[256] =  {
 0   ,
 1   ,
 2   ,
 3   ,
 4   ,
 5   ,
 6   ,
 7   ,
 8   ,
 9   ,
 10  ,
 11  ,
 12  ,
 13  ,
 14  ,
 15  ,
 16  ,
 17  ,
 18  ,
 19  ,
 20  ,
 21  ,
 22  ,
 23  ,
 24  ,
 25  ,
 26  ,
 27  ,
 28  ,
 29  ,
 30  ,
 31  ,
 32  ,
 33  ,
 34  ,
 35  ,
 36  ,
 37  ,
 38  ,
 39  ,
 40  ,
 41  ,
 42  ,
 43  ,
 44  ,
 45  ,
 46  ,
 0   ,
 48  ,
 49  ,
 50  ,
 51  ,
 52  ,
 53  ,
 54  ,
 55  ,
 56  ,
 57  ,
 58  ,
 59  ,
 60  ,
 61  ,
 62  ,
 63  ,
 64  ,
 65  ,
 66  ,
 67  ,
 68  ,
 69  ,
 70  ,
 71  ,
 72  ,
 73  ,
 74  ,
 75  ,
 76  ,
 77  ,
 78  ,
 79  ,
 80  ,
 81  ,
 82  ,
 83  ,
 84  ,
 85  ,
 86  ,
 86  ,
 88  ,
 85  ,
 90  ,
 91  ,
 92  ,
 93  ,
 94  ,
 95  ,
 96  ,
 65  ,
 66  ,
 67  ,
 68  ,
 69  ,
 70  ,
 71  ,
 72  ,
 73  ,
 74  ,
 75  ,
 76  ,
 77  ,
 78  ,
 79  ,
 80  ,
 81  ,
 82  ,
 83  ,
 84  ,
 85  ,
 86  ,
 86  ,
 88  ,
 85  ,
 90  ,
 123 ,
 124 ,
 125 ,
 126 ,
 127 ,
 127 ,
 127 ,
 130 ,
 70  ,
 132 ,
 133 ,
 134 ,
 135 ,
 127 ,
 137 ,
 83  ,
 139 ,
 140 ,
 127 ,
 127 ,
 127 ,
 127 ,
 145 ,
 146 ,
 147 ,
 148 ,
 149 ,
 150 ,
 150 ,
 152 ,
 153 ,
 83  ,
 155 ,
 140 ,
 127 ,
 127 ,
 85  ,
 160 ,
 161 ,
 162 ,
 163 ,
 164 ,
 165 ,
 166 ,
 167 ,
 168 ,
 169 ,
 65  ,
 171 ,
 172 ,
 150 ,
 174 ,
 175 ,
 176 ,
 177 ,
 50  ,
 51  ,
 180 ,
 181 ,
 182 ,
 183 ,
 184 ,
 49  ,
 79  ,
 187 ,
 188 ,
 189 ,
 190 ,
 191 ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 67  ,
 69  ,
 69  ,
 69  ,
 69  ,
 73  ,
 73  ,
 73  ,
 73  ,
 68  ,
 78  ,
 79  ,
 79  ,
 79  ,
 79  ,
 79  ,
 215 ,
 79  ,
 85  ,
 85  ,
 85  ,
 85  ,
 85  ,
 222 ,
 223 ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 65  ,
 67  ,
 69  ,
 69  ,
 69  ,
 69  ,
 73  ,
 73  ,
 73  ,
 73  ,
 68  ,
 78  ,
 79  ,
 79  ,
 79  ,
 79  ,
 79  ,
 247 ,
 79  ,
 85  ,
 85  ,
 85  ,
 85  ,
 85  ,
 222 ,
 85
};
#pragma code_seg()

#pragma code_seg(CS_CP10000)
ALLOC_CODE(BYTE) g_rgbCodePage10000[256] =  {
 0  ,
 1  ,
 2  ,
 3  ,
 4  ,
 5  ,
 6  ,
 7  ,
 8  ,
 9  ,
 10 ,
 11 ,
 12 ,
 13 ,
 14 ,
 15 ,
 16 ,
 17 ,
 18 ,
 19 ,
 20 ,
 21 ,
 22 ,
 23 ,
 24 ,
 25 ,
 26 ,
 27 ,
 28 ,
 29 ,
 30 ,
 31 ,
 32 ,
 33 ,
 34 ,
 35 ,
 36 ,
 37 ,
 38 ,
 39 ,
 40 ,
 41 ,
 42 ,
 43 ,
 44 ,
 45 ,
 46 ,
 0  ,
 48 ,
 49 ,
 50 ,
 51 ,
 52 ,
 53 ,
 54 ,
 55 ,
 56 ,
 57 ,
 58 ,
 59 ,
 60 ,
 61 ,
 62 ,
 63 ,
 64 ,
 65 ,
 66 ,
 67 ,
 68 ,
 69 ,
 70 ,
 71 ,
 72 ,
 73 ,
 74 ,
 75 ,
 76 ,
 77 ,
 78 ,
 79 ,
 80 ,
 81 ,
 82 ,
 83 ,
 84 ,
 85 ,
 86 ,
 86 ,
 88 ,
 85 ,
 90 ,
 91 ,
 92 ,
 93 ,
 94 ,
 95 ,
 96 ,
 65 ,
 66 ,
 67 ,
 68 ,
 69 ,
 70 ,
 71 ,
 72 ,
 73 ,
 74 ,
 75 ,
 76 ,
 77 ,
 78 ,
 79 ,
 80 ,
 81 ,
 82 ,
 83 ,
 84 ,
 85 ,
 86 ,
 86 ,
 88 ,
 85 ,
 90 ,
 123,
 124,
 125,
 126,
 127,
 65 ,
 65 ,
 67 ,
 69 ,
 78 ,
 79 ,
 85 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 171,
 172,
 173,
 65 ,
 79 ,
 176,
 177,
 178,
 179,
 180,
 181,
 182,
 183,
 184,
 185,
 186,
 65 ,
 79 ,
 189,
 65 ,
 79 ,
 192,
 193,
 194,
 195,
 70 ,
 197,
 198,
 199,
 200,
 201,
 202,
 65 ,
 65 ,
 79 ,
 206,
 206,
 208,
 208,
 210,
 211,
 212,
 213,
 214,
 215,
 85 ,
 85 ,
 218,
 219,
 220,
 221,
 63 ,
 63 ,
 224,
 225,
 226,
 227,
 228,
 65 ,
 69 ,
 65 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 79 ,
 79 ,
 63 ,
 79 ,
 85 ,
 85 ,
 85 ,
 73 ,
 127,
 247,
 127,
 249,
 250,
 251,
 63 ,
 253,
 254,
 127
};
#pragma code_seg()

#pragma code_seg(CS_CP1250)
ALLOC_CODE(BYTE) g_rgbCodePage1250[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
0  ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
127,
127,
127,
130,
127,
132,
133,
134,
135,
127,
137,
83 ,
139,
83 ,
84 ,
90 ,
90 ,
127,
145,
146,
147,
148,
149,
150,
150,
127,
153,
83 ,
155,
83 ,
84 ,
90 ,
90 ,
160,
127,
162,
76 ,
164,
65 ,
166,
167,
168,
169,
83 ,
171,
172,
150,
174,
90 ,
176,
177,
178,
76 ,
180,
181,
182,
183,
184,
65 ,
83 ,
187,
76 ,
189,
76 ,
90 ,
82 ,
65 ,
65 ,
65 ,
65 ,
76 ,
67 ,
67 ,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
68 ,
208,
78 ,
78 ,
79 ,
79 ,
79 ,
79 ,
215,
82 ,
85 ,
85 ,
85 ,
85 ,
89 ,
84 ,
223,
82 ,
65 ,
65 ,
65 ,
65 ,
76 ,
67 ,
67 ,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
68 ,
208,
78 ,
78 ,
79 ,
79 ,
79 ,
79 ,
247,
82 ,
85 ,
85 ,
85 ,
85 ,
89 ,
84 ,
255
};
#pragma code_seg()

#pragma code_seg(CS_CP1251)
ALLOC_CODE(BYTE) g_rgbCodePage1251[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
0  ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
127,
127,
127,
130,
70 ,
132,
133,
134,
135,
127,
137,
83 ,
139,
140,
127,
127,
127,
127,
145,
146,
147,
148,
149,
150,
150,
152,
153,
83 ,
155,
140,
127,
127,
89 ,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
150,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
215,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
223,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
247,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
89
};
#pragma code_seg()

#pragma code_seg(CS_CP10029)
ALLOC_CODE(BYTE) g_rgbCodePage10029[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
0  ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
127,
65 ,
65 ,
65 ,
69 ,
65 ,
79 ,
85 ,
65 ,
65 ,
67 ,
65 ,
67 ,
67 ,
67 ,
69 ,
90 ,
90 ,
68 ,
73 ,
68 ,
69 ,
69 ,
69 ,
79 ,
69 ,
79 ,
79 ,
79 ,
85 ,
69 ,
69 ,
85 ,
160,
161,
69 ,
163,
164,
165,
166,
167,
168,
169,
170,
69 ,
172,
173,
71 ,
73 ,
73 ,
73 ,
178,
179,
73 ,
75 ,
182,
183,
76 ,
76 ,
76 ,
76 ,
76 ,
76 ,
76 ,
78 ,
78 ,
78 ,
194,
195,
78 ,
78 ,
198,
199,
200,
201,
202,
78 ,
79 ,
79 ,
79 ,
79 ,
208,
208,
210,
211,
212,
213,
214,
215,
79 ,
82 ,
82 ,
82 ,
220,
221,
82 ,
82 ,
82 ,
83 ,
226,
227,
83 ,
83 ,
83 ,
65 ,
84 ,
84 ,
73 ,
90 ,
90 ,
85 ,
79 ,
79 ,
85 ,
85 ,
85 ,
85 ,
85 ,
85 ,
85 ,
85 ,
89 ,
89 ,
75 ,
90 ,
76 ,
76 ,
71 ,
255
};
#pragma code_seg()

#pragma code_seg(CS_CP10007)
ALLOC_CODE(BYTE) g_rgbCodePage10007[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
0  ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
127,
128,
129,
130,
131,
132,
133,
134,
135,
136,
137,
138,
139,
140,
141,
142,
143,
144,
145,
146,
147,
148,
149,
150,
151,
152,
153,
154,
155,
156,
157,
158,
159,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
170,
171,
171,
173,
174,
174,
176,
177,
178,
179,
167,
181,
182,
183,
184,
184,
186,
186,
188,
188,
190,
190,
183,
193,
194,
195,
70 ,
197,
198,
199,
200,
201,
202,
203,
203,
205,
205,
193,
208,
208,
210,
211,
212,
213,
214,
215,
216,
216,
218,
218,
220,
221,
221,
159,
128,
129,
130,
131,
132,
133,
134,
135,
136,
137,
138,
139,
140,
141,
142,
143,
144,
145,
146,
147,
148,
149,
150,
151,
152,
153,
154,
155,
156,
157,
158,
159
};
#pragma code_seg()

#pragma code_seg(CS_CPWGREEK)
ALLOC_CODE(BYTE) g_rgbCodePageWGreek[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
0  ,
139,
0  ,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
0  ,
153,
0  ,
155,
0  ,
0  ,
0  ,
0  ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
0  ,
171,
172,
45 ,
174,
45 ,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
185,
186,
187,
188,
189,
190,
191,
186,
162,
194,
195,
196,
184,
198,
185,
200,
186,
202,
203,
204,
205,
206,
188,
208,
209,
0  ,
211,
212,
190,
214,
215,
216,
191,
186,
190,
162,
184,
185,
186,
190,
162,
194,
195,
196,
184,
198,
185,
200,
186,
202,
203,
204,
205,
206,
188,
208,
209,
211,
211,
212,
190,
214,
215,
216,
191,
186,
190,
188,
190,
191,
0
};
#pragma code_seg()

#pragma code_seg(CS_CPWICELAND)
ALLOC_CODE(BYTE) g_rgbCodePageWIceland[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
193,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
201,
69 ,
69 ,
73 ,
205,
73 ,
73 ,
208,
78 ,
79 ,
211,
79 ,
79 ,
214,
215,
214,
85 ,
218,
85 ,
85 ,
221,
222,
223,
65 ,
193,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
201,
69 ,
69 ,
73 ,
205,
73 ,
73 ,
208,
78 ,
79 ,
211,
79 ,
79 ,
214,
247,
214,
85 ,
218,
85 ,
85 ,
221,
222,
89
};
#pragma code_seg()

#pragma code_seg(CS_CPWTURKISH)
ALLOC_CODE(BYTE) g_rgbCodePageWTurkish[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
199,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
215,
79 ,
85 ,
85 ,
85 ,
220,
221,
222,
223,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
199,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
247,
79 ,
85 ,
85 ,
85 ,
220,
221,
222,
89
};
#pragma code_seg()

#pragma code_seg(CS_CPWNORWEGIAN)
ALLOC_CODE(BYTE) g_rgbCodePageWNorwegian[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
196,
197,
196,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
68 ,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
215,
214,
85 ,
85 ,
85 ,
89 ,
89 ,
222,
223,
65 ,
65 ,
65 ,
65 ,
196,
197,
196,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
68 ,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
247,
214,
85 ,
85 ,
85 ,
89 ,
89 ,
222,
89
};
#pragma code_seg()

#pragma code_seg(CS_CPWENGIRELAND)
ALLOC_CODE(BYTE) g_rgbCodePageWEngIreland[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
215,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
223,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
247,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
89
};

#pragma code_seg()

#pragma code_seg(CS_CPMGREEK)
ALLOC_CODE(BYTE) g_rgbCodePageMGreek[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
65 ,
49 ,
50 ,
69 ,
51 ,
79 ,
85 ,
135,
65 ,
65 ,
65 ,
0  ,
140,
67 ,
69 ,
69 ,
69 ,
69 ,
146,
147,
73 ,
73 ,
150,
151,
152,
79 ,
79 ,
155,
63 ,
85 ,
85 ,
85 ,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
170,
171,
172,
173,
174,
175,
176,
177,
178,
179,
180,
181,
182,
183,
184,
171,
186,
187,
188,
189,
190,
191,
176,
193,
194,
195,
196,
197,
198,
199,
200,
201,
9  ,
189,
204,
176,
182,
207,
45 ,
45 ,
210,
211,
212,
213,
214,
184,
171,
195,
189,
182,
184,
171,
195,
191,
189,
176,
181,
190,
162,
182,
188,
161,
184,
171,
165,
186,
164,
187,
193,
195,
166,
191,
196,
170,
198,
163,
191,
170,
204,
189,
183,
171,
189,
171,
189,
63
};

#pragma code_seg()

#pragma code_seg(CS_CPMNORWEGIAN)
ALLOC_CODE(BYTE) g_rgbCodePageMNorwegian[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
33 ,
34 ,
35 ,
36 ,
37 ,
38 ,
39 ,
40 ,
41 ,
42 ,
43 ,
44 ,
45 ,
46 ,
47 ,
48 ,
49 ,
50 ,
51 ,
52 ,
53 ,
54 ,
55 ,
56 ,
57 ,
58 ,
59 ,
60 ,
61 ,
62 ,
63 ,
64 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
91 ,
92 ,
93 ,
94 ,
95 ,
96 ,
65 ,
66 ,
67 ,
68 ,
69 ,
70 ,
71 ,
72 ,
73 ,
74 ,
75 ,
76 ,
77 ,
78 ,
79 ,
80 ,
81 ,
82 ,
83 ,
84 ,
85 ,
86 ,
87 ,
88 ,
89 ,
90 ,
123,
124,
125,
126,
0  ,
128,
129,
67 ,
69 ,
78 ,
133,
89 ,
65 ,
65 ,
65 ,
128,
65 ,
129,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
78 ,
79 ,
79 ,
79 ,
133,
79 ,
85 ,
85 ,
85 ,
89 ,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
170,
171,
172,
173,
128,
133,
176,
177,
178,
179,
180,
181,
182,
183,
184,
185,
186,
65 ,
79 ,
189,
128,
133,
192,
193,
194,
195,
70 ,
197,
198,
199,
200,
201,
9  ,
65 ,
65 ,
79 ,
206,
206,
45 ,
45 ,
210,
211,
212,
213,
214,
215,
89 ,
89 ,
218,
219,
220,
221,
63 ,
63 ,
224,
225,
226,
227,
228,
65 ,
69 ,
65 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
79 ,
79 ,
63 ,
79 ,
85 ,
85 ,
85 ,
73 ,
0  ,
247,
0  ,
249,
250,
251,
63 ,
253,
254,
0
};

#pragma code_seg()

#pragma code_seg(CS_CPMENGIRELAND)
ALLOC_CODE(BYTE) g_rgbCodePageMEngIreland[256] =  {
0  ,
1  ,
2  ,
3  ,
4  ,
5  ,
6  ,
7  ,
8  ,
9  ,
10 ,
11 ,
12 ,
13 ,
14 ,
15 ,
16 ,
17 ,
18 ,
19 ,
20 ,
21 ,
22 ,
23 ,
24 ,
25 ,
26 ,
27 ,
28 ,
29 ,
30 ,
31 ,
32 ,
 33 ,
 34 ,
 35 ,
 36 ,
 37 ,
 38 ,
 39 ,
 40 ,
 41 ,
 42 ,
 43 ,
 44 ,
 45 ,
 46 ,
 47 ,
 48 ,
 49 ,
 50 ,
 51 ,
 52 ,
 53 ,
 54 ,
 55 ,
 56 ,
 57 ,
 58 ,
 59 ,
 60 ,
 61 ,
 62 ,
 63 ,
 64 ,
 65 ,
 66 ,
 67 ,
 68 ,
 69 ,
 70 ,
 71 ,
 72 ,
 73 ,
 74 ,
 75 ,
 76 ,
 77 ,
 78 ,
 79 ,
 80 ,
 81 ,
 82 ,
 83 ,
 84 ,
 85 ,
 86 ,
 87 ,
 88 ,
 89 ,
 90 ,
 91 ,
 92 ,
 93 ,
 94 ,
 95 ,
 96 ,
 65 ,
 66 ,
 67 ,
 68 ,
 69 ,
 70 ,
 71 ,
 72 ,
 73 ,
 74 ,
 75 ,
 76 ,
 77 ,
 78 ,
 79 ,
 80 ,
 81 ,
 82 ,
 83 ,
 84 ,
 85 ,
 86 ,
 87 ,
 88 ,
 89 ,
 90 ,
 123,
 124,
 125,
 126,
 0  ,
 65 ,
 65 ,
 67 ,
 69 ,
 78 ,
 79 ,
 85 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 171,
 172,
 173,
 174,
 79 ,
 176,
 177,
 178,
 179,
 180,
 181,
 182,
 183,
 184,
 185,
 186,
 65 ,
 79 ,
 189,
 174,
 79 ,
 192,
 193,
 194,
 195,
 70 ,
 197,
 198,
 199,
 200,
 201,
 9  ,
 65 ,
 65 ,
 79 ,
 206,
 206,
 45 ,
 45 ,
 210,
 211,
 212,
 213,
 214,
 215,
 89 ,
 89 ,
 218,
 219,
 220,
 221,
 63 ,
 63 ,
 224,
 225,
 226,
 227,
 228,
 65 ,
 69 ,
 65 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 79 ,
 79 ,
 63 ,
 79 ,
 85 ,
 85 ,
 85 ,
 73 ,
 0  ,
 247,
 0  ,
 249,
 250,
 251,
 63 ,
 253,
 254,
 0
};
#pragma code_seg()

// BIDI stuff
#pragma code_seg(CS_CP1255)
ALLOC_CODE(BYTE) g_rgbCodePage1255[256] =  {
 0, 
 1, 
 2, 
 3, 
 4, 
 5, 
 6, 
 7, 
 8, 
 9, 
 10, 
 11, 
 12, 
 13, 
 14, 
 15, 
 16, 
 17, 
 18, 
 19, 
 20, 
 21, 
 22, 
 23, 
 24, 
 25, 
 26, 
 27, 
 28, 
 29, 
 30, 
 31, 
 32, 
 33, 
 34, 
 35, 
 36, 
 37, 
 38, 
 39, 
 40, 
 41, 
 42, 
 43, 
 44, 
 45, 
 46, 
 47, 
 48, 
 49, 
 50, 
 51, 
 52, 
 53, 
 54, 
 55, 
 56, 
 57, 
 58, 
 59, 
 60, 
 61, 
 62, 
 63, 
 64, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 91, 
 92, 
 93, 
 94, 
 95, 
 96, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 123, 
 124, 
 125, 
 126, 
 127, 
 128, 
 129, 
 130, 
 70, 
 132, 
 133, 
 134, 
 135, 
 94, 
 137, 
 138, 
 139, 
 140, 
 141, 
 142, 
 143, 
 144, 
 145, 
 146, 
 147, 
 148, 
 149, 
 150, 
 151, 
 152, 
 153, 
 154, 
 155, 
 156, 
 157, 
 158, 
 159, 
 160, 
 161, 
 162, 
 163, 
 164, 
 165, 
 166, 
 167, 
 168, 
 169, 
 170, 
 171, 
 172, 
 173, 
 174, 
 175, 
 176, 
 177, 
 50, 
 51, 
 180, 
 181, 
 182, 
 183, 
 184, 
 49, 
 186, 
 187, 
 188, 
 189, 
 190, 
 191, 
 192, 
 193, 
 194, 
 195, 
 196, 
 197, 
 198, 
 199, 
 200, 
 201, 
 202, 
 203, 
 204, 
 205, 
 206, 
 207, 
 208, 
 209, 
 210, 
 211, 
 212, 
 213, 
 214, 
 215, 
 216, 
 217, 
 218, 
 219, 
 220, 
 221, 
 222, 
 223, 
 224, 
 225, 
 226, 
 227, 
 228, 
 229, 
 230, 
 231, 
 232, 
 233, 
 234, 
 234, 
 236, 
 237, 
 237, 
 239, 
 239, 
 241, 
 242, 
 243, 
 243, 
 245, 
 245, 
 247, 
 248, 
 249, 
 250, 
 251, 
 252, 
 0, 
 0, 
 255
};
#pragma code_seg()

#pragma code_seg(CS_CP1256)
ALLOC_CODE(BYTE) g_rgbCodePage1256[256] =  {
 0, 
 1, 
 2, 
 3, 
 4, 
 5, 
 6, 
 7, 
 8, 
 9, 
 10, 
 11, 
 12, 
 13, 
 14, 
 15, 
 16, 
 17, 
 18, 
 19, 
 20, 
 21, 
 22, 
 23, 
 24, 
 25, 
 26, 
 27, 
 28, 
 29, 
 30, 
 31, 
 32, 
 33, 
 34, 
 35, 
 36, 
 37, 
 38, 
 39, 
 40, 
 41, 
 42, 
 43, 
 44, 
 45, 
 46, 
 47, 
 48, 
 49, 
 50, 
 51, 
 52, 
 53, 
 54, 
 55, 
 56, 
 57, 
 58, 
 59, 
 60, 
 61, 
 62, 
 63, 
 64, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 91, 
 92, 
 93, 
 94, 
 95, 
 96, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 123, 
 124, 
 125, 
 126, 
 127, 
 128, 
 129, 
 130, 
 70, 
 132, 
 133, 
 134, 
 135, 
 94, 
 137, 
 138, 
 139, 
 140, 
 141, 
 142, 
 143, 
 144, 
 145, 
 146, 
 147, 
 148, 
 149, 
 150, 
 151, 
 152, 
 153, 
 154, 
 155, 
 140, 
 157, 
 0, 
 159, 
 160, 
 161, 
 162, 
 163, 
 164, 
 165, 
 166, 
 167, 
 168, 
 169, 
 170, 
 171, 
 172, 
 173, 
 174, 
 175, 
 176, 
 177, 
 50, 
 51, 
 180, 
 181, 
 182, 
 183, 
 184, 
 49, 
 186, 
 187, 
 188, 
 189, 
 190, 
 191, 
 192, 
 193, 
 194, 
 193, 
 193, 
 193, 
 193, 
 199, 
 200, 
 201, 
 201, 
 203, 
 204, 
 205, 
 206, 
 207, 
 208, 
 209, 
 210, 
 211, 
 212, 
 213, 
 214, 
 215, 
 216, 
 217, 
 218, 
 219, 
 0, 
 221, 
 222, 
 223, 
 65, 
 225, 
 65, 
 227, 
 228, 
 229, 
 230, 
 67, 
 69, 
 69, 
 69, 
 69, 
 236, 
 236, 
 73, 
 73, 
 240, 
 241, 
 242, 
 243, 
 79, 
 245, 
 246, 
 247, 
 248, 
 85, 
 250, 
 85, 
 85, 
 0, 
 0, 
 255
};
#pragma code_seg()

#pragma code_seg(CS_CP10004)
ALLOC_CODE(BYTE) g_rgbCodePage10004[256] =  {
 0, 
 1, 
 2, 
 3, 
 4, 
 5, 
 6, 
 7, 
 8, 
 9, 
 10, 
 11, 
 12, 
 13, 
 14, 
 15, 
 16, 
 17, 
 18, 
 19, 
 20, 
 21, 
 22, 
 23, 
 24, 
 25, 
 26, 
 27, 
 28, 
 29, 
 30, 
 31, 
 32, 
 33, 
 34, 
 35, 
 36, 
 37, 
 38, 
 39, 
 40, 
 41, 
 42, 
 43, 
 44, 
 45, 
 46, 
 47, 
 48, 
 49, 
 50, 
 51, 
 52, 
 53, 
 54, 
 55, 
 56, 
 57, 
 58, 
 59, 
 60, 
 61, 
 62, 
 63, 
 64, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 91, 
 92, 
 93, 
 94, 
 95, 
 96, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 123, 
 124, 
 125, 
 126, 
 127, 
 65, 
 129, 
 67, 
 69, 
 78, 
 79, 
 85, 
 65, 
 65, 
 65, 
 65, 
 139, 
 140, 
 67, 
 69, 
 69, 
 69, 
 69, 
 73, 
 147, 
 73, 
 73, 
 78, 
 79, 
 152, 
 79, 
 79, 
 155, 
 85, 
 85, 
 85, 
 85, 
 160, 
 161, 
 162, 
 163, 
 164, 
 165, 
 166, 
 167, 
 168, 
 169, 
 170, 
 171, 
 172, 
 173, 
 174, 
 175, 
 176, 
 177, 
 178, 
 179, 
 180, 
 181, 
 182, 
 183, 
 184, 
 185, 
 186, 
 187, 
 188, 
 189, 
 190, 
 191, 
 192, 
 193, 
 194, 
 193, 
 193, 
 193, 
 193, 
 199, 
 200, 
 201, 
 201, 
 203, 
 204, 
 205, 
 206, 
 207, 
 208, 
 209, 
 210, 
 211, 
 212, 
 213, 
 214, 
 215, 
 216, 
 217, 
 218, 
 219, 
 220, 
 221, 
 222, 
 223, 
 0, 
 225, 
 226, 
 227, 
 228, 
 229, 
 230, 
 231, 
 232, 
 233, 
 233, 
 235, 
 236, 
 237, 
 238, 
 239, 
 240, 
 241, 
 242, 
 243, 
 244, 
 245, 
 246, 
 247, 
 248, 
 249, 
 250, 
 251, 
 252, 
 253, 
 254, 
 255, 
};
#pragma code_seg()

#pragma code_seg(CS_CP10005)
ALLOC_CODE(BYTE) g_rgbCodePage10005[256] =  {
 0, 
 1, 
 2, 
 3, 
 4, 
 5, 
 6, 
 7, 
 8, 
 9, 
 10, 
 11, 
 12, 
 13, 
 14, 
 15, 
 16, 
 17, 
 18, 
 19, 
 20, 
 21, 
 22, 
 23, 
 24, 
 25, 
 26, 
 27, 
 28, 
 29, 
 30, 
 31, 
 32, 
 33, 
 34, 
 35, 
 36, 
 37, 
 38, 
 39, 
 40, 
 41, 
 42, 
 43, 
 44, 
 45, 
 46, 
 47, 
 48, 
 49, 
 50, 
 51, 
 52, 
 53, 
 54, 
 55, 
 56, 
 57, 
 58, 
 59, 
 60, 
 61, 
 62, 
 63, 
 64, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 91, 
 92, 
 93, 
 94, 
 95, 
 96, 
 65, 
 66, 
 67, 
 68, 
 69, 
 70, 
 71, 
 72, 
 73, 
 74, 
 75, 
 76, 
 77, 
 78, 
 79, 
 80, 
 81, 
 82, 
 83, 
 84, 
 85, 
 86, 
 87, 
 88, 
 89, 
 90, 
 123, 
 124, 
 125, 
 126, 
 127, 
 65, 
 65, 
 67, 
 69, 
 78, 
 79, 
 85, 
 65, 
 65, 
 65, 
 65, 
 65, 
 65, 
 67, 
 69, 
 69, 
 69, 
 69, 
 73, 
 73, 
 73, 
 73, 
 78, 
 79, 
 79, 
 79, 
 79, 
 79, 
 85, 
 85, 
 85, 
 85, 
 160, 
 161, 
 162, 
 163, 
 164, 
 165, 
 166, 
 167, 
 168, 
 169, 
 170, 
 171, 
 172, 
 173, 
 174, 
 175, 
 176, 
 177, 
 178, 
 179, 
 180, 
 181, 
 182, 
 183, 
 184, 
 185, 
 186, 
 187, 
 188, 
 189, 
 190, 
 191, 
 192, 
 193, 
 194, 
 195, 
 196, 
 197, 
 198, 
 199, 
 200, 
 201, 
 202, 
 203, 
 204, 
 205, 
 206, 
 207, 
 208, 
 209, 
 210, 
 211, 
 212, 
 213, 
 214, 
 215, 
 216, 
 217, 
 218, 
 219, 
 220, 
 221, 
 222, 
 223, 
 224, 
 225, 
 226, 
 227, 
 228, 
 229, 
 230, 
 231, 
 232, 
 233, 
 234, 
 234, 
 236, 
 237, 
 237, 
 239, 
 239, 
 241, 
 242, 
 243, 
 243, 
 245, 
 245, 
 247, 
 248, 
 249, 
 250, 
 251, 
 252, 
 253, 
 254, 
 255, 
};
#pragma code_seg()
 
 

// This is 2-D table of Code Page Vs. System.
// To get the address of the mapping table to use; Use
// pb = g_prgHashTbl[CodePage][Syskind];
//
ALLOC_CODE(BYTE *)g_prgHashTbl[15][2] = {
    {NULL,		NULL},
    {g_rgbCodePage1252, g_rgbCodePage10000}, // English/European
    {g_rgbCodePage1250, g_rgbCodePage10029},
    {g_rgbCodePage1251, g_rgbCodePage10007},
    {NULL,		NULL},		     // JAPNESE_CP
    {NULL,		NULL},		     // KOREAN_CP
    {NULL,		NULL},		     // CHINESES_CP
    {NULL,		NULL},		     // CHINESET_CP

    // MAC: UNDONE: Tables of Turkish and Icelandic is not available on MAC.
    {g_rgbCodePageWGreek,      g_rgbCodePageMGreek	    }, // GREEK
    {g_rgbCodePageWIceland,    g_rgbCodePageWIceland	    }, // ICELANDIC
    {g_rgbCodePageWTurkish,    g_rgbCodePageWTurkish	    }, // TURKISH
    {g_rgbCodePageWNorwegian,  g_rgbCodePageMNorwegian	    }, // NORWEIGIAN
    {g_rgbCodePageWEngIreland, g_rgbCodePageMEngIreland     }, // ENGLISH(IRELAND)
    {g_rgbCodePage1256, g_rgbCodePage10004}, // Arabic
    {g_rgbCodePage1255, g_rgbCodePage10005}  // Hebrew
};

// Following are the code page "codes" for the dbcs world.
#define JAPNESE_CP    0x4
#define KOREAN_CP     0x5
#define CHINESES_CP   0x6
#define CHINESET_CP   0x7

#define GREEK_CP      0x8
#define ICELANDIC_CP  0x9
#define TURKISH_CP    0xa
#define NORWEIGIAN_CP 0xb
#define IRELAND_CP    0xc
#define ARABIC_CP     0xd
#define HEBREW_CP     0xe


/***
*PUBLIC LHashValOfName
*Purpose: This API calculates the hash value depending on the system  and
*	  the lcid that was passed in.
*
*Implementation Notes:
*    Use the table pointed by  pbHashTable to map each character in the
*    szName.
*
*
*Entry:
*	xsz: String whose hash value has to be calculated
*
*Exit:
*   ULONG :   hash value of the passed in string
*
*    31      23  19   15		0
*   |-------|---|----|-------------------|
*   |UNUSED |CP |Sys | wHashVal 	 |
*   |_______|___|____|___________________|
*	       ^   ^
*	       |   |
*	CodePageId Encoding bits for System kind
*
*
*
***********************************************************************/
#pragma optimize("t", on)

#if OE_WIN32
STDAPI_(unsigned long) LHashValOfNameSysW(SYSKIND syskind,
					 LCID lcid,
					 LPCOLESTR szName)
{
    // NOTE: We only bother to translate the first 255 chars of the name.
    // NOTE: This should sill get a pretty good hash function, without
    // NOTE: having burn more stack space, or (worse) alloc & free memory
    // NOTE: on this call.
    char szNameA[255+1];

    if (szName == NULL)		// can't return E_INVALIDARG here, so just
	return 0;		// don't crash

    // convert szName from Unicode to Ansi.
    NoAssertRetail (WideCharToMultiByte(CP_ACP,
					0,
					szName,
					-1,
					szNameA,
					255+1,
					NULL,
					NULL) != 0, "");
    szNameA[255] = '\0';	// ensure null-terminated (in case truncated).

    return LHashValOfNameSysA(syskind, lcid, szNameA);
}
#endif  //OE_WIN32

STDAPI_(unsigned long) LHashValOfNameSysA(SYSKIND syskind,
					 LCID lcid,
					 const char FAR* szName)
{
    ULONG ulHash;		// hash value which accumulates hash
    USHORT uIndex;
    BYTE *pbHashTable;
    USHORT usSystem, usCodePage;
    ULONG uEncoding;

    if (szName == NULL)		// can't return E_INVALIDARG here, so just
	return 0;		// don't crash

    // If DBCS then defer to LHashValOfDBCSName
    if (IsDBCS(lcid))
      return LHashValOfDBCSName(syskind, lcid, szName);

    // Deternine the hash table that should be used.
    //
    // Find the index corresponding to the system passed in.
    usSystem = (syskind == SYS_MAC) ? 1 : 0;

    // Determine the Code page.
    if (lcid == 0x409)
	usCodePage = 1; // English is most likely.
    else switch (lcid) {
      default:
	if (LOBYTE(lcid) == 0x01 || lcid == 0x0429) {
	  usCodePage = ARABIC_CP; break;
	}
	usCodePage = 1; break;  // default gets US English (1252)
      case 0x0419:
	usCodePage = 3; break;
      case 0x0408:
	usCodePage = GREEK_CP; break;
      case 0x040f:
	usCodePage = ICELANDIC_CP; break;
      case 0x041f:
	usCodePage = TURKISH_CP; break;
      case 0x0814:
	usCodePage = NORWEIGIAN_CP; break;
      case 0x1809:
	usCodePage = IRELAND_CP; break;
      case 0x040d:
	usCodePage = HEBREW_CP; break;
      case 0x0405:
      case 0x040e:
      case 0x0415:
      case 0x041b:
	usCodePage = 2; break;
      }




    pbHashTable =  g_prgHashTbl[usCodePage][usSystem];

    uEncoding = (ULONG) (((ULONG) ((usCodePage << 4) | usSystem)) << 16);


    // Algorithm: compute 32-bit value by multipling by 37 and adding
    // in the character.  Ignore all overflows.  Then convert to
    // a 16-bit number by moding by 65599 (a prime).

    ulHash = 0xDeadBee;    // arbitrary starting number


    for (;*szName; szName++) {

      // use  pbHashTable[] to map characters
      //

      uIndex = ((BYTE) *szName) & 0x00ff;

      ulHash = ulHash * 37 + pbHashTable[uIndex];

    } // for

    ulHash = ulHash % 65599;

    return (ULONG) (uEncoding | (ulHash & 0x0000FFFF));
}



/***
*LHashValOfDBCSName
*Purpose: Calculates the hash value for strings from DBCS language.
*
*Implementation Notes:
*    Use the table pointed by  pbHashTable to map each character in the
*    szName.
*
*
*Entry:
*   xsz: String whose hash value has to be calculated
*
*Exit:
*   ULONG :   hash value of the passed in string
*
*
***********************************************************************/
#pragma code_seg(CS_DBCSHASH)
unsigned long LHashValOfDBCSName(SYSKIND syskind, LCID lcid, const char FAR* szName)
{
    ULONG ulHash;		// hash value which accumulates hash
    ACHAR xch;
    USHORT uIndex;
    BYTE *pbHashTable;
    USHORT usSystem, usCodePage;
    ULONG uEncoding;

    // UNDONE : for MAC
    //
    // For DBCS characters < 255, the mapping rules are same as CODE page for
    // US english.
    //
    pbHashTable = g_rgbCodePage1252;

    // Find the index corresponding to the system passed in.
    usSystem = (syskind == SYS_MAC) ? 1 : 0;

    // Get the code page.
    switch (PRIMARYLANGID(lcid)) {
    case LANG_JAPANESE:
      usCodePage = JAPNESE_CP;
      break;
    case LANG_KOREAN:
      usCodePage = KOREAN_CP;
      break;
    case LANG_CHINESE:
      switch (SUBLANGID(lcid)) {
      case SUBLANG_CHINESE_TRADITIONAL:
        usCodePage = CHINESET_CP;
        break;
      case SUBLANG_CHINESE_SIMPLIFIED:
        usCodePage = CHINESES_CP;
        break;
      default:
        // We default to chinese_simplified sub-lang.
        usCodePage = CHINESES_CP;
        break;
      } 
      break;
    default:
      DebHalt("bad lcid");
    }

    uEncoding = (ULONG) (((ULONG) ((usCodePage << 4) | usSystem)) << 16);

    // Algorithm: compute 32-bit value by multipling by 37 and adding
    // in the character.  Ignore all overflows.  Then convert to
    // a 16-bit number by moding by 65599 (a prime).

    ulHash = 0xDeadBee;    // arbitrary starting number

    for (; (xch = xpch((LPSTR)szName)) != '\0'; szName = xstrinc(szName)) {

      // use  pbHashTable[] to map characters
      //

      // If the next character is a double byte character then first map it.
      if ((USHORT)xch > 0xff) {
        xch = MapDBChar(xch, lcid);

        // process the high byte.
        // shift the upper byte down to get the index.
        if (uIndex = (((USHORT) xch & 0xff00) >> 8)) {
	  ulHash = ulHash * 37 + pbHashTable[uIndex];
        }
      }

      // process the low byte.
      uIndex = ((USHORT) xch) & 0x00ff;

      ulHash = ulHash * 37 + pbHashTable[uIndex];

    } // for

    ulHash = ulHash % 65599;

    return (ULONG) (uEncoding | (ulHash & 0x0000FFFF));

}
#pragma optimize("t", )



ALLOC_CODE(ACHAR) m_rgJapneseExcepMap[][2] = {
0x829F,      0x00A7,
0x82A0,      0x00B1,
0x82A1,      0x00A8,
0x82A2,      0x00B2,
0x82A3,      0x00A9,
0x82A4,      0x00B3,
0x82A5,      0x00AA,
0x82A6,      0x00B4,
0x82A7,      0x00AB,
0x82A8,      0x00B5,
0x82A9,      0x00B6,
0x82AA,      0xB6DE,
0x82AB,      0x00B7,
0x82AC,      0xB7DE,
0x82AD,      0x00B8,
0x82AE,      0xB8DE,
0x82AF,      0x00B9,
0x82B0,      0xB9DE,
0x82B1,      0x00BA,
0x82B2,      0xBADE,
0x82B3,      0x00BB,
0x82B4,      0xBBDE,
0x82B5,      0x00BC,
0x82B6,      0xBCDE,
0x82B7,      0x00BD,
0x82B8,      0xBDDE,
0x82B9,      0x00BE,
0x82BA,      0xBEDE,
0x82BB,      0x00BF,
0x82BC,      0xBFDE,
0x82BD,      0x00C0,
0x82BE,      0xC0DE,
0x82BF,      0x00C1,
0x82C0,      0xC1DE,
0x82C1,      0x00AF,
0x82C2,      0x00C2,
0x82C3,      0xC2DE,
0x82C4,      0x00C3,
0x82C5,      0xC3DE,
0x82C6,      0x00C4,
0x82C7,      0xC4DE,
0x82C8,      0x00C5,
0x82C9,      0x00C6,
0x82CA,      0x00C7,
0x82CB,      0x00C8,
0x82CC,      0x00C9,
0x82CD,      0x00CA,
0x82CE,      0xCADE,
0x82CF,      0xCADF,
0x82D0,      0x00CB,
0x82D1,      0xCBDE,
0x82D2,      0xCBDF,
0x82D3,      0x00CC,
0x82D4,      0xCCDE,
0x82D5,      0xCCDF,
0x82D6,      0x00CD,
0x82D7,      0xCDDE,
0x82D8,      0xCDDF,
0x82D9,      0x00CE,
0x82DA,      0xCEDE,
0x82DB,      0xCEDF,
0x82DC,      0x00CF,
0x82DD,      0x00D0,
0x82DE,      0x00D1,
0x82DF,      0x00D2,
0x82E0,      0x00D3,
0x82E1,      0x00AC,
0x82E2,      0x00D4,
0x82E3,      0x00AD,
0x82E4,      0x00D5,
0x82E5,      0x00AE,
0x82E6,      0x00D6,
0x82E7,      0x00D7,
0x82E8,      0x00D8,
0x82E9,      0x00D9,
0x82EA,      0x00DA,
0x82EB,      0x00DB,
0x82EC,      0x838E,
0x82ED,      0x00DC,
0x82EE,      0x8390,
0x82EF,      0x8391,
0x82F0,      0x00A6,
0x82F1,      0x00DD,

0x8340,      0x00A7,
0x8341,      0x00B1,
0x8342,      0x00A8,
0x8343,      0x00B2,
0x8344,      0x00A9,
0x8345,      0x00B3,
0x8346,      0x00AA,
0x8347,      0x00B4,
0x8348,      0x00AB,
0x8349,      0x00B5,
0x834A,      0x00B6,
0x834B,      0xB6DE,
0x834C,      0x00B7,
0x834D,      0xB7DE,
0x834E,      0x00B8,
0x834F,      0xB8DE,
0x8350,      0x00B9,
0x8351,      0xB9DE,
0x8352,      0x00BA,
0x8353,      0xBADE,
0x8354,      0x00BB,
0x8355,      0xBBDE,
0x8356,      0x00BC,
0x8357,      0xBCDE,
0x8358,      0x00BD,
0x8359,      0xBDDE,
0x835A,      0x00BE,
0x835B,      0xBEDE,
0x835C,      0x00BF,
0x835D,      0xBFDE,
0x835E,      0x00C0,
0x835F,      0xC0DE,
0x8360,      0x00C1,
0x8361,      0xC1DE,
0x8362,      0x00AF,
0x8363,      0x00C2,
0x8364,      0xC2DE,
0x8365,      0x00C3,
0x8366,      0xC3DE,
0x8367,      0x00C4,
0x8368,      0xC4DE,
0x8369,      0x00C5,
0x836A,      0x00C6,
0x836B,      0x00C7,
0x836C,      0x00C8,
0x836D,      0x00C9,
0x836E,      0x00CA,
0x836F,      0xCADE,
0x8370,      0xCADF,
0x8371,      0x00CB,
0x8372,      0xCBDE,
0x8373,      0xCBDF,
0x8374,      0x00CC,
0x8375,      0xCCDE,
0x8376,      0xCCDF,
0x8377,      0x00CD,
0x8378,      0xCDDE,
0x8379,      0xCDDF,
0x837A,      0x00CE,
0x837B,      0xCEDE,
0x837C,      0xCEDF,
0x837D,      0x00CF,
0x837E,      0x00D0,
0x8380,      0x00D1,
0x8381,      0x00D2,
0x8382,      0x00D3,
0x8383,      0x00AC,
0x8384,      0x00D4,
0x8385,      0x00AD,
0x8386,      0x00D5,
0x8387,      0x00AE,
0x8388,      0x00D6,
0x8389,      0x00D7,
0x838A,      0x00D8,
0x838B,      0x00D9,
0x838C,      0x00DA,
0x838D,      0x00DB,
0x838E,      0x838E,
0x838F,      0x00DC,
0x8390,      0x8390,
0x8391,      0x8391,
0x8392,      0x00A6,
0x8393,      0x00DD,
0x8394,      0xB3DE,
0x8395,      0x8395,
0x8396,      0x8396,

0x824F,      0x30,
0x8250,      0x31,
0x8251,      0x32,
0x8252,      0x33,
0x8253,      0x34,
0x8254,      0x35,
0x8255,      0x36,
0x8256,      0x37,
0x8257,      0x38,
0x8258,      0x39,
0x8260,      0x41,
0x8261,      0x42,
0x8262,      0x43,
0x8263,      0x44,
0x8264,      0x45,
0x8265,      0x46,
0x8266,      0x47,
0x8267,      0x48,
0x8268,      0x49,
0x8269,      0x4A,
0x826A,      0x4B,
0x826B,      0x4C,
0x826C,      0x4D,
0x826D,      0x4E,
0x826E,      0x4F,
0x826F,      0x50,
0x8270,      0x51,
0x8271,      0x52,
0x8272,      0x53,
0x8273,      0x54,
0x8274,      0x55,
0x8275,      0x56,
0x8276,      0x57,
0x8277,      0x58,
0x8278,      0x59,
0x8279,      0x5A,
0x8151,      0x5F,
0x8281,      0x61,
0x8282,      0x62,
0x8283,      0x63,
0x8284,      0x64,
0x8285,      0x65,
0x8286,      0x66,
0x8287,      0x67,
0x8288,      0x68,
0x8289,      0x69,
0x828A,      0x6A,
0x828B,      0x6B,
0x828C,      0x6C,
0x828D,      0x6D,
0x828E,      0x6E,
0x828F,      0x6F,
0x8290,      0x70,
0x8291,      0x71,
0x8292,      0x72,
0x8293,      0x73,
0x8294,      0x74,
0x8295,      0x75,
0x8296,      0x76,
0x8297,      0x77,
0x8298,      0x78,
0x8299,      0x79,
0x829A,      0x7A,
0x0000,      0x00
};      

ALLOC_CODE(ACHAR) m_rgKoreaExcepMap[][2] = {
0xA3B0,      0x30,
0xA3B1,      0x31,
0xA3B2,      0x32,
0xA3B3,      0x33,
0xA3B4,      0x34,
0xA3B5,      0x35,
0xA3B6,      0x36,
0xA3B7,      0x37,
0xA3B8,      0x38,
0xA3B9,      0x39,
0xA3C1,      0x41,
0xA3C2,      0x42,
0xA3C3,      0x43,
0xA3C4,      0x44,
0xA3C5,      0x45,
0xA3C6,      0x46,
0xA3C7,      0x47,
0xA3C8,      0x48,
0xA3C9,      0x49,
0xA3CA,      0x4A,
0xA3CB,      0x4B,
0xA3CC,      0x4C,
0xA3CD,      0x4D,
0xA3CE,      0x4E,
0xA3CF,      0x4F,
0xA3D0,      0x50,
0xA3D1,      0x51,
0xA3D2,      0x52,
0xA3D3,      0x53,
0xA3D4,      0x54,
0xA3D5,      0x55,
0xA3D6,      0x56,
0xA3D7,      0x57,
0xA3D8,      0x58,
0xA3D9,      0x59,
0xA3DA,      0x5A,
0xA3DF,      0x5F,
0xA3E1,      0x61,
0xA3E2,      0x62,
0xA3E3,      0x63,
0xA3E4,      0x64,
0xA3E5,      0x65,
0xA3E6,      0x66,
0xA3E7,      0x67,
0xA3E8,      0x68,
0xA3E9,      0x69,
0xA3EA,      0x6A,
0xA3EB,      0x6B,
0xA3EC,      0x6C,
0xA3ED,      0x6D,
0xA3EE,      0x6E,
0xA3EF,      0x6F,
0xA3F0,      0x70,
0xA3F1,      0x71,
0xA3F2,      0x72,
0xA3F3,      0x73,
0xA3F4,      0x74,
0xA3F5,      0x75,
0xA3F6,      0x76,
0xA3F7,      0x77,
0xA3F8,      0x78,
0xA3F9,      0x79,
0xA3FA,      0x7A,
0x0000,      0x00
};


ALLOC_CODE(ACHAR) m_rgChinaTExcepMap[][2] = {
0xA2AF,      0x30,
0xA2B0,      0x31,
0xA2B1,      0x32,
0xA2B2,      0x33,
0xA2B3,      0x34,
0xA2B4,      0x35,
0xA2B5,      0x36,
0xA2B6,      0x37,
0xA2B7,      0x38,
0xA2B8,      0x39,
0xA2CF,      0x41,
0xA2D0,      0x42,
0xA2D1,      0x43,
0xA2D2,      0x44,
0xA2D3,      0x45,
0xA2D4,      0x46,
0xA2D5,      0x47,
0xA2D6,      0x48,
0xA2D7,      0x49,
0xA2D8,      0x4A,
0xA2D9,      0x4B,
0xA2DA,      0x4C,
0xA2DB,      0x4D,
0xA2DC,      0x4E,
0xA2DD,      0x4F,
0xA2DE,      0x50,
0xA2DF,      0x51,
0xA2E0,      0x52,
0xA2E1,      0x53,
0xA2E2,      0x54,
0xA2E3,      0x55,
0xA2E4,      0x56,
0xA2E5,      0x57,
0xA2E6,      0x58,
0xA2E7,      0x59,
0xA2E8,      0x5A,
0xA1C5,      0x5F,
0xA2E9,      0x61,
0xA2EA,      0x62,
0xA2EB,      0x63,
0xA2EC,      0x64,
0xA2ED,      0x65,
0xA2EE,      0x66,
0xA2EF,      0x67,
0xA2F0,      0x68,
0xA2F1,      0x69,
0xA2F2,      0x6A,
0xA2F3,      0x6B,
0xA2F4,      0x6C,
0xA2F5,      0x6D,
0xA2F6,      0x6E,
0xA2F7,      0x6F,
0xA2F8,      0x70,
0xA2F9,      0x71,
0xA2FA,      0x72,
0xA2FB,      0x73,
0xA2FC,      0x74,
0xA2FD,      0x75,
0xA2FE,      0x76,
0xA340,      0x77,
0xA341,      0x78,
0xA342,      0x79,
0xA343,      0x7A,
};

ALLOC_CODE(ACHAR) m_rgChinaSExcepMap[][2] = {
0xA3B0,      0x30,
0xA3B1,      0x31,
0xA3B2,      0x32,
0xA3B3,      0x33,
0xA3B4,      0x34,
0xA3B5,      0x35,
0xA3B6,      0x36,
0xA3B7,      0x37,
0xA3B8,      0x38,
0xA3B9,      0x39,
0xA3C1,      0x41,
0xA3C2,      0x42,
0xA3C3,      0x43,
0xA3C4,      0x44,
0xA3C5,      0x45,
0xA3C6,      0x46,
0xA3C7,      0x47,
0xA3C8,      0x48,
0xA3C9,      0x49,
0xA3CA,      0x4A,
0xA3CB,      0x4B,
0xA3CC,      0x4C,
0xA3CD,      0x4D,
0xA3CE,      0x4E,
0xA3CF,      0x4F,
0xA3D0,      0x50,
0xA3D1,      0x51,
0xA3D2,      0x52,
0xA3D3,      0x53,
0xA3D4,      0x54,
0xA3D5,      0x55,
0xA3D6,      0x56,
0xA3D7,      0x57,
0xA3D8,      0x58,
0xA3D9,      0x59,
0xA3DA,      0x5A,
0xA3DF,      0x5F,
0xA3E1,      0x61,
0xA3E2,      0x62,
0xA3E3,      0x63,
0xA3E4,      0x64,
0xA3E5,      0x65,
0xA3E6,      0x66,
0xA3E7,      0x67,
0xA3E8,      0x68,
0xA3E9,      0x69,
0xA3EA,      0x6A,
0xA3EB,      0x6B,
0xA3EC,      0x6C,
0xA3ED,      0x6D,
0xA3EE,      0x6E,
0xA3EF,      0x6F,
0xA3F0,      0x70,
0xA3F1,      0x71,
0xA3F2,      0x72,
0xA3F3,      0x73,
0xA3F4,      0x74,
0xA3F5,      0x75,
0xA3F6,      0x76,
0xA3F7,      0x77,
0xA3F8,      0x78,
0xA3F9,      0x79,
0xA3FA,      0x7A,
};


/***
*MapDBChar
*Purpose: Maps a double byte character to a (1 or 2) single byte character
*
*Implementation Notes:
*    Use the exception tables above for the appropriate language.
*
*Entry:
*   xch: Character to be mapped
*   lcid: LCID of language in use
*
*Exit:
*   returns mapped character
*
*CONSIDER:  there are some optimizations that could be done here
*
***********************************************************************/
ACHAR MapDBChar(ACHAR xch, LCID lcid)
{
    ACHAR * pchException;

    DebAssert(0xff00 & (USHORT)xch, " should a double byte character ");

    // Do language specific mapping.
    switch (PRIMARYLANGID(lcid)) {
    case LANG_JAPANESE:
      // Japan
      pchException = m_rgJapneseExcepMap[0];
      break;
    case LANG_KOREAN:
      // Korea
      pchException = m_rgKoreaExcepMap[0];
      break;
    case LANG_CHINESE:
      switch (SUBLANGID(lcid)) {
      case SUBLANG_CHINESE_TRADITIONAL:
        // China(T)
        pchException = m_rgChinaTExcepMap[0];
        break;
      case SUBLANG_CHINESE_SIMPLIFIED:
        // China(S)
        pchException = m_rgChinaSExcepMap[0];
        break;
      default:
        DebHalt("bad lcid");
      }
      break;
    default:
      DebHalt("bad lcid");
    }

    for (; *pchException != 0x0000 && *pchException != xch; pchException += 2);
    // end of for loop

    if (*pchException == 0x0000)
      return xch;
    else
      return *(pchException+1);

}
#pragma code_seg()


// *****  Start case mapping tables

/////////////////////////////////////////////////////////////////////////
//  The following tables/Code is used to generate the tables for accent and
//  and case insensitive comparision.
//



/******
*        0x0409 // These uses the base table it self.
*        0x0407,
*        0x040c,
*        0x0809,
*        0x0c09,
*        0x1009,
*        0x1409,
*        0x0413,
*        0x0813,
*        0x080c,
*        0x0c0c,
*        0x100c,
*        0x0807,
*        0x0c07,
*        0x0410,
*        0x0810,
*        0x0416,
*        0x0816,
*        0x0419  // Russian(in different code page):
*                // but the comparision table is same as US english.
*
*        0x040b, // these uses g_rgbExcepTblWin1035 as execption table
*        0x041d,
*
*
*        0x0406, // these uses g_rgbExcepTblWin1030 as execption table
*        0x0414,
*
*
*        0x040a, // these uses g_rgbExcepTblWin1034 as execption table
*        0x080a,
*        0x0c0a,
*
*
* These falls in a different  Code Page (WIN:1240, MAC:10029)
*        0x0405, These uses the base table it self.
*
*        0x040e,
*
*        0x0415,
*
*        0x041b,
***************************************************************************/

// Partial base Tables.  The partial tables are used to build the CodePage
// and sysytem kind dependent base table. We the full blown table when we
// load the typelib.dll
//

// In the partial base tables we do not have the first 128 characters.
// We construct the first 128 characters in the following way:
// From 0 to 96 we  have identity (index == the entry).
// From 97 to 122 we have 65 to 90 (note this is 'a' to 'z' that naps to 'A' to 'Z').
// From 123 to 127 we again have identity.
//
#pragma code_seg(CS_CORE2)
ALLOC_CODE(BYTE) g_rgbPartialBaseTbl1252[128] = {
 127,
 127,
 130,
 70 ,
 132,
 133,
 134,
 135,
 127,
 137,
 83 ,
 139,
 140,
 127,
 127,
 127,
 127,
 145,
 146,
 147,
 148,
 149,
 150,
 150,
 152,
 153,
 83 ,
 155,
 140,
 127,
 127,
 89 ,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 65 ,
 171,
 172,
 150,
 174,
 175,
 176,
 177,
 50 ,
 51 ,
 180,
 181,
 182,
 183,
 184,
 49 ,
 79 ,
 187,
 188,
 189,
 190,
 191,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 198,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 208,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 215,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 89 ,
 222,
 223,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 198,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 208,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 247,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 89 ,
 222,
 89  
};

ALLOC_CODE(BYTE) g_rgbPartialBaseTbl1250[128] = {
 127,
 127,
 130,
 127,
 132,
 133,
 134,
 135,
 127,
 137,
 138,
 139,
 83 ,
 84 ,
 142,
 90 ,
 127,
 145,
 146,
 147,
 148,
 149,
 150,
 150,
 127,
 153,
 138,
 155,
 83 ,
 84 ,
 142,
 90 ,
 160,
 127,
 162,
 76 ,
 164,
 65 ,
 166,
 167,
 168,
 169,
 83 ,
 171,
 172,
 150,
 174,
 90 ,
 176,
 177,
 178,
 76 ,
 180,
 181,
 182,
 183,
 184,
 65 ,
 83 ,
 187,
 76 ,
 189,
 76 ,
 90 ,
 82 ,
 65 ,
 65 ,
 65 ,
 65 ,
 76 ,
 67 ,
 67 ,
 200,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 68 ,
 208,
 78 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 215,
 216,
 85 ,
 85 ,
 85 ,
 85 ,
 89 ,
 84 ,
 223,
 82 ,
 65 ,
 65 ,
 65 ,
 65 ,
 76 ,
 67 ,
 67 ,
 200,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 68 ,
 208,
 78 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 247,
 216,
 85 ,
 85 ,
 85 ,
 85 ,
 89 ,
 84 ,
 255
};

ALLOC_CODE(BYTE) g_rgbPartialBaseTbl10000[128] = {
 65 ,
 65 ,
 67 ,
 69 ,
 78 ,
 79 ,
 85 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 171,
 172,
 173,
 174,
 79 ,
 176,
 177,
 178,
 179,
 180,
 181,
 182,
 183,
 184,
 185,
 186,
 65 ,
 79 ,
 189,
 174,
 79 ,
 192,
 193,
 194,
 195,
 70 ,
 197,
 198,
 199,
 200,
 201,
 202,
 65 ,
 65 ,
 79 ,
 206,
 206,
 208,
 208,
 210,
 211,
 212,
 213,
 214,
 215,
 89 ,
 89 ,
 218,
 219,
 220,
 221,
 63 ,
 63 ,
 224,
 225,
 226,
 227,
 228,
 65 ,
 69 ,
 65 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 79 ,
 79 ,
 63 ,
 79 ,
 85 ,
 85 ,
 85 ,
 73 ,
 127,
 247,
 127,
 249,
 250,
 251,
 63 ,
 253,
 254,
 127

};

ALLOC_CODE(BYTE) g_rgbPartialBaseTbl10029[128] = {
 65 ,
 65 ,
 65 ,
 69 ,
 65 ,
 79 ,
 85 ,
 65 ,
 65 ,
 137,
 65 ,
 137,
 67 ,
 67 ,
 69 ,
 90 ,
 90 ,
 68 ,
 73 ,
 68 ,
 69 ,
 69 ,
 69 ,
 79 ,
 69 ,
 79 ,
 79 ,
 79 ,
 85 ,
 69 ,
 69 ,
 85 ,
 160,
 161,
 69 ,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 69 ,
 172,
 173,
 71 ,
 73 ,
 73 ,
 73 ,
 178,
 179,
 73 ,
 75 ,
 182,
 183,
 76 ,
 76 ,
 76 ,
 76 ,
 76 ,
 76 ,
 76 ,
 78 ,
 78 ,
 78 ,
 194,
 195,
 78 ,
 78 ,
 198,
 199,
 200,
 201,
 202,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 208,
 208,
 210,
 211,
 212,
 213,
 214,
 215,
 79 ,
 82 ,
 82 ,
 219,
 220,
 221,
 219,
 82 ,
 82 ,
 225,
 226,
 227,
 225,
 83 ,
 83 ,
 65 ,
 84 ,
 84 ,
 73 ,
 235,
 235,
 85 ,
 79 ,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 85 ,
 85 ,
 85 ,
 85 ,
 89 ,
 89 ,
 75 ,
 90 ,
 76 ,
 76 ,
 71 ,
 255


};

ALLOC_CODE(BYTE) g_rgbPartialBaseTbl10007[128] = {
 128,
 129,
 130,
 131,
 132,
 133,
 134,
 135,
 136,
 137,
 138,
 139,
 140,
 141,
 142,
 143,
 144,
 145,
 146,
 147,
 148,
 149,
 150,
 151,
 152,
 153,
 154,
 155,
 156,
 157,
 158,
 159,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 171,
 171,
 173,
 174,
 174,
 176,
 177,
 178,
 179,
 167,
 181,
 182,
 183,
 184,
 184,
 186,
 186,
 188,
 188,
 190,
 190,
 183,
 193,
 194,
 195,
 70 ,
 197,
 198,
 199,
 200,
 201,
 202,
 203,
 203,
 205,
 205,
 193,
 208,
 208,
 210,
 211,
 212,
 213,
 214,
 215,
 216,
 216,
 218,
 218,
 220,
 221,
 221,
 159,
 128,
 129,
 130,
 131,
 132,
 133,
 134,
 135,
 136,
 137,
 138,
 139,
 140,
 141,
 142,
 143,
 144,
 145,
 146,
 147,
 148,
 149,
 150,
 151,
 152,
 153,
 154,
 155,
 156,
 157,
 158,
 159 

};


// Now we use  g_rgbCodePageWGreek to build the table.
#if 0
ALLOC_CODE(BYTE) g_rgbPartialBaseTblWGreek[128] =  {
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
0  ,
139,
0  ,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
0  ,
153,
0  ,
155,
0  ,
0  ,
0  ,
0  ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
0  ,
171,
172,
45 ,
174,
45 ,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
185,
186,
187,
188,
189,
190,
191,
186,
162,
194,
195,
196,
184,
198,
185,
200,
186,
202,
203,
204,
205,
206,
188,
208,
209,
0  ,
211,
212,
190,
214,
215,
216,
191,
186,
190,
162,
184,
185,
186,
190,
162,
194,
195,
196,
184,
198,
185,
200,
186,
202,
203,
204,
205,
206,
188,
208,
209,
211,
211,
212,
190,
214,
215,
216,
191,
186,
190,
188,
190,
191,
0
};
#endif //0

// Now we use g_rgbCodePageWIceland to build the table.
#if 0
ALLOC_CODE(BYTE) g_rgbPartialBaseTblWIceland[128] =  {
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
193,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
201,
69 ,
69 ,
73 ,
205,
73 ,
73 ,
208,
78 ,
79 ,
211,
79 ,
79 ,
214,
215,
214,
85 ,
218,
85 ,
85 ,
221,
222,
223,
65 ,
193,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
201,
69 ,
69 ,
73 ,
205,
73 ,
73 ,
208,
78 ,
79 ,
211,
79 ,
79 ,
214,
247,
214,
85 ,
218,
85 ,
85 ,
221,
222,
89
};

#endif //0


ALLOC_CODE(BYTE) g_rgbPartialBaseTblWTurkish[128] =  {
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
199,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
215,
79 ,
85 ,
85 ,
85 ,
220,
221,
222,
223,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
199,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
247,
79 ,
85 ,
85 ,
85 ,
220,
221,
222,
89
};

ALLOC_CODE(BYTE) g_rgbPartialBaseTblWNorwegian[128] =  {
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
196,
197,
196,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
68 ,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
215,
214,
85 ,
85 ,
85 ,
89 ,
89 ,
222,
223,
65 ,
65 ,
65 ,
65 ,
196,
197,
196,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
68 ,
78 ,
79 ,
79 ,
79 ,
79 ,
214,
247,
214,
85 ,
85 ,
85 ,
89 ,
89 ,
222,
89
};

ALLOC_CODE(BYTE) g_rgbPartialBaseTblWEngIreland[128] =  {
0  ,
0  ,
130,
70 ,
132,
133,
134,
135,
0  ,
137,
83 ,
139,
140,
0  ,
0  ,
0  ,
0  ,
145,
146,
147,
148,
149,
45 ,
45 ,
152,
153,
83 ,
155,
140,
0  ,
0  ,
89 ,
9  ,
161,
162,
163,
164,
165,
166,
167,
168,
169,
65 ,
171,
172,
45 ,
174,
175,
176,
177,
50 ,
51 ,
180,
181,
182,
183,
184,
49 ,
79 ,
187,
188,
189,
190,
191,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
215,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
223,
65 ,
65 ,
65 ,
65 ,
65 ,
65 ,
198,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
208,
78 ,
79 ,
79 ,
79 ,
79 ,
79 ,
247,
79 ,
85 ,
85 ,
85 ,
85 ,
89 ,
222,
89
};


// Now we use  g_rgbCodePageMGreek to build the table.
#if 0
ALLOC_CODE(BYTE) g_rgbPartialBaseTblMGreek[128] =  {
65 ,
49 ,
50 ,
69 ,
51 ,
79 ,
85 ,
135,
65 ,
65 ,
65 ,
0  ,
140,
67 ,
69 ,
69 ,
69 ,
69 ,
146,
147,
73 ,
73 ,
150,
151,
152,
79 ,
79 ,
155,
63 ,
85 ,
85 ,
85 ,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
170,
171,
172,
173,
174,
175,
176,
177,
178,
179,
180,
181,
182,
183,
184,
171,
186,
187,
188,
189,
190,
191,
176,
193,
194,
195,
196,
197,
198,
199,
200,
201,
9  ,
189,
204,
176,
182,
207,
45 ,
45 ,
210,
211,
212,
213,
214,
184,
171,
195,
189,
182,
184,
171,
195,
191,
189,
176,
181,
190,
162,
182,
188,
161,
184,
171,
165,
186,
164,
187,
193,
195,
166,
191,
196,
170,
198,
163,
191,
170,
204,
189,
183,
171,
189,
171,
189,
63
};
#endif // 0

ALLOC_CODE(BYTE) g_rgbPartialBaseTblMNorwegian[128] =  {
128,
129,
67 ,
69 ,
78 ,
133,
89 ,
65 ,
65 ,
65 ,
128,
65 ,
129,
67 ,
69 ,
69 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
78 ,
79 ,
79 ,
79 ,
133,
79 ,
85 ,
85 ,
85 ,
89 ,
160,
161,
162,
163,
164,
165,
166,
167,
168,
169,
170,
171,
172,
173,
128,
133,
176,
177,
178,
179,
180,
181,
182,
183,
184,
185,
186,
65 ,
79 ,
189,
128,
133,
192,
193,
194,
195,
70 ,
197,
198,
199,
200,
201,
9  ,
65 ,
65 ,
79 ,
206,
206,
45 ,
45 ,
210,
211,
212,
213,
214,
215,
89 ,
89 ,
218,
219,
220,
221,
63 ,
63 ,
224,
225,
226,
227,
228,
65 ,
69 ,
65 ,
69 ,
69 ,
73 ,
73 ,
73 ,
73 ,
79 ,
79 ,
63 ,
79 ,
85 ,
85 ,
85 ,
73 ,
0  ,
247,
0  ,
249,
250,
251,
63 ,
253,
254,
0
};

ALLOC_CODE(BYTE) g_rgbPartialBaseTblMEngIreland[128] =  {
 65 ,
 65 ,
 67 ,
 69 ,
 78 ,
 79 ,
 85 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 65 ,
 67 ,
 69 ,
 69 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 78 ,
 79 ,
 79 ,
 79 ,
 79 ,
 79 ,
 85 ,
 85 ,
 85 ,
 85 ,
 160,
 161,
 162,
 163,
 164,
 165,
 166,
 167,
 168,
 169,
 170,
 171,
 172,
 173,
 174,
 79 ,
 176,
 177,
 178,
 179,
 180,
 181,
 182,
 183,
 184,
 185,
 186,
 65 ,
 79 ,
 189,
 174,
 79 ,
 192,
 193,
 194,
 195,
 70 ,
 197,
 198,
 199,
 200,
 201,
 9  ,
 65 ,
 65 ,
 79 ,
 206,
 206,
 45 ,
 45 ,
 210,
 211,
 212,
 213,
 214,
 215,
 89 ,
 89 ,
 218,
 219,
 220,
 221,
 63 ,
 63 ,
 224,
 225,
 226,
 227,
 228,
 65 ,
 69 ,
 65 ,
 69 ,
 69 ,
 73 ,
 73 ,
 73 ,
 73 ,
 79 ,
 79 ,
 63 ,
 79 ,
 85 ,
 85 ,
 85 ,
 73 ,
 0  ,
 247,
 0  ,
 249,
 250,
 251,
 63 ,
 253,
 254,
 0
};



// Table that is used to generate the base table for each code page for
// WIN and MAC.
ALLOC_CODE(BYTE*) g_prgbPartialBaseTbl[11][2] = {
    // WIN             // MAC
    {NULL,              NULL},
    {g_rgbPartialBaseTbl1252, g_rgbPartialBaseTbl10000},
    {g_rgbPartialBaseTbl1250, g_rgbPartialBaseTbl10029},
    {g_rgbPartialBaseTbl1252, g_rgbPartialBaseTbl10007},

    // MAC: UNDONE: Tables of Turkish and Icelandic is not available on MAC.
// UNDONE: they are now available. Add them.
    {g_rgbCodePageWGreek+128,          g_rgbCodePageMGreek+128      }, // GREEK
    {g_rgbCodePageWIceland+128,        g_rgbCodePageWIceland+128    }, // ICELANDIC
    {g_rgbPartialBaseTblWTurkish,    g_rgbPartialBaseTblWTurkish    }, // TURKISH
    {g_rgbPartialBaseTblWNorwegian,  g_rgbPartialBaseTblMNorwegian  }, // NORWEIGIAN
    {g_rgbPartialBaseTblWEngIreland, g_rgbPartialBaseTblMEngIreland }, // ENGLISH(IRELAND)
    {g_rgbCodePage1256+128,          g_rgbCodePage10004+128}, // Arabic
    {g_rgbCodePage1255+128,          g_rgbCodePage10005+128}  // Hebrew

};


#define GREEK_CP_Partial      0x4
#define ICELANDIC_CP_Partial  0x5
#define TURKISH_CP_Partial    0x6
#define NORWEIGIAN_CP_Partial 0x7
#define IRELAND_CP_Partial    0x8
#define ARABIC_CP_Partial    0x9
#define HEBREW_CP_Partial    0xa


ALLOC_CODE(BYTE) g_rgbExcepTblWin1030[][2] = {
    {196, 196},
    {197, 197},
    {198, 196},
    {208, 68 },
    {214, 214},
    {216, 214},
    {220, 89 },
    {228, 196},
    {229, 197},
    {230, 196},
    {240, 68 },
    {246, 214},
    {248, 214},
    {252, 89 },
    {0,   0}
};

ALLOC_CODE(BYTE) g_rgbExcepTblMac1030[][2] = {
    {128, 128},
    {129, 129},
    {133, 133},
    {134,  89},
    {138, 128},
    {140, 129},
    {154, 133},
    {159, 89 },
    {174, 128},
    {175, 133},
    {190, 128},
    {191, 133},
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblWin1034[][2] = {
    {209, 209},
    {241, 209},
    {0,   0}
};

ALLOC_CODE(BYTE) g_rgbExcepTblMac1034[][2] = {
    {132, 132},
    {150, 132},
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblWin1035[][2] = {
    {87,  86 },
    {119, 86 },
    {196, 196},
    {197, 197},
    {208, 68 },
    {214, 214},
    {216, 214},
    {220, 89 },
    {228, 196},
    {229, 197},
    {240, 68 },
    {246, 214},
    {248, 214},
    {252, 89 },
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblMac1035[][2] = {
    {87,  86 },
    {119, 86 },
    {128, 128},
    {129, 129},
    {133, 133},
    {134, 89 },
    {138, 128},
    {140, 129},
    {154, 133},
    {159, 89 },
    {175, 133},
    {191, 133},
    {0,   0}
};



ALLOC_CODE(BYTE) g_rgbExcepTblWin1038[][2] = {
    {138, 83 },
    {142, 90 },
    {154, 83 },
    {158, 90 },
    {200, 67 },
    {213, 213},
    {214, 213},
    {216, 82 },
    {218, 218},
    {219, 219},
    {220, 219},
    {232, 67 },
    {245, 213},
    {246, 213},
    {248, 82 },
    {250, 218},
    {251, 219},
    {252, 219},
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblMac1038[][2] = {
    {133, 133},
    {134, 134},
    {137, 67 },
    {139, 67 },
    {154, 133},
    {156, 156},
    {159, 134},
    {204, 133},
    {206, 133},
    {219, 82 },
    {222, 82 },
    {225, 83 },
    {228, 83 },
    {235, 90 },
    {236, 90 },
    {242, 156},
    {244, 134},
    {245, 134},
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblWin1045[][2] = {
    {138, 83 },
    {140, 140},
    {142, 90 },
    {143, 143},
    {154, 83 },
    {156, 140},
    {158, 90 },
    {159, 143},
    {163, 163},
    {165, 165},
    {175, 175},
    {179, 163},
    {185, 165},
    {191, 175},
    {198, 198},
    {200, 67 },
    {202, 202},
    {209, 209},
    {211, 211},
    {216, 82 },
    {230, 198},
    {232, 67 },
    {241, 209},
    {243, 211},
    {248, 82 },
    {0,   0}
};

ALLOC_CODE(BYTE) g_rgbExcepTblMac1045[][2] = {
    {132, 132},
    {136, 132},
    {137, 67 },
    {139, 67 },
    {140, 140},
    {140, 140},
    {143, 143},
    {144, 143},
    {151, 151},
    {162, 162},
    {171, 162},
    {184, 184},
    {193, 193},
    {196, 193},
    {219, 82 },
    {222, 82 },
    {225, 83 },
    {228, 83 },
    {229, 229},
    {230, 229},
    {235, 90 },
    {236, 90 },
    {238, 151},
    {251, 251},
    {252, 184},
    {0,   0}
};


ALLOC_CODE(BYTE) g_rgbExcepTblWin1051[][2] = {
    {196, 196},
    {212, 212},
    {228, 196},
    {244, 212},
    {0,   0}
};

ALLOC_CODE(BYTE) g_rgbExcepTblMac1051[][2] = {
    {138, 138},
    {153, 153},
    {239, 153},
    {0,   0}
};

ALLOC_CODE(BYTE) g_rgbExcepTblNew[][2] = {
    {127, 0},
    {0,   0}
};


struct LcidToExcep {
    LCID m_lcid;
    VOID *pvExcepTbl;
};

// Table containing pointers to the exception table for both windows and
// Mac platform.
ALLOC_CODE(struct LcidToExcep) g_rgLcidToExcepTbl[][2] = {
     // WIN                             // MAC
   {  {1030,  g_rgbExcepTblWin1030},     {1030,  g_rgbExcepTblMac1030},  },
   {  {1044,  g_rgbExcepTblWin1030},     {1044,  g_rgbExcepTblMac1030},  },
   {  {1035,  g_rgbExcepTblWin1035},     {1035,  g_rgbExcepTblMac1035},  },
   {  {1053,  g_rgbExcepTblWin1035},     {1053,  g_rgbExcepTblMac1035},  },
   {  {1034,  g_rgbExcepTblWin1034},     {1034,  g_rgbExcepTblMac1034},  },
   {  {2058,  g_rgbExcepTblWin1034},     {2058,  g_rgbExcepTblMac1034},  },
   {  {3082,  g_rgbExcepTblWin1034},     {3082,  g_rgbExcepTblMac1034},  },
   {  {0x403, g_rgbExcepTblWin1034},     {0x403,  g_rgbExcepTblMac1034},  },
   {  {1038,  g_rgbExcepTblWin1038},     {1038,  g_rgbExcepTblMac1038},  },
   {  {1045,  g_rgbExcepTblWin1045},     {1045,  g_rgbExcepTblMac1045},  },
   {  {1051,  g_rgbExcepTblWin1051},     {1051,  g_rgbExcepTblMac1051},  },

   {  {1032,  g_rgbExcepTblNew},   {1032,  g_rgbExcepTblNew},  },
   {  {1039,  g_rgbExcepTblNew},   {1039,  g_rgbExcepTblNew},  },
   {  {1055,  g_rgbExcepTblNew},   {1055,  g_rgbExcepTblNew},  },
   {  {2068,  g_rgbExcepTblNew},   {2068,  g_rgbExcepTblNew},  },
   {  {6153,  g_rgbExcepTblNew},   {6153,  g_rgbExcepTblNew},  },

   {  {0, NULL},                          {0, NULL}                       }
};



/***
*VOID  GetInsensitiveCompTbl()
*Purpose: This function returns a table that can be used for accent and case
*         insensitive comparision.
*IMPLEMENTATION : There are 2 steps to build the table. First build the base
*                 table for the code page that contains the passed in lcid.
*                 Then see if there is any exception table for that code page.
*                 If there is then fix up the exception cases.
*
*      From 0 to 96 we  have identity (index == the entry).
*      From 97 to 122 we have 65 to 90 (note this is 'a' to 'z' that naps to 'A' to 'Z').
*      From 123 to 127 we again have identity.
*
*Inputs:
*   lcid : locale for which that table is required.
*   syskind : the system for which the table is required.
*
*Outputs:
*   rgchTbl : This table is initialized. This can be used to do accent and
*             case insensitive comparision.
*
*
****************************************************************************/
VOID GetInsensitiveCompTbl(LCID lcid, SYSKIND syskind, XCHAR *rgchTbl)
{
    BYTE *pbPartialTable;
    // BYTE *pbrgExcepTable[2];
    BYTE *pbExcepTable;
    USHORT usSystem, usCodePage;
    UINT i;

    DebAssert(rgchTbl != NULL, " Bad Argument ");

    // In retail version, return if we have a bad argument.
    if (rgchTbl == NULL)
      return;


    // Deternine the partail base table to use to get the generate the
    // base table.
    //
    // Find the index corresponding to the system passed in.
    usSystem = (syskind == SYS_MAC) ? 1 : 0;

    // Determine the Code page.
    if (lcid == 0x0419) {
      usCodePage = 3;
    }
    else
      if (lcid == 0x0408) {
  usCodePage = GREEK_CP_Partial;
      }
      else
  if (lcid == 0x040f) {
    usCodePage = ICELANDIC_CP_Partial;
  }
  else
    if (lcid == 0x041f) {
      usCodePage = TURKISH_CP_Partial;
    }
    else
      if (lcid == 0x0814) {
        usCodePage = NORWEIGIAN_CP_Partial;
      }
      else
        if (lcid == 0x1809) {
    usCodePage = IRELAND_CP_Partial;
        }
        else
    if (lcid == 0x0405 ||
      lcid == 0x040e ||
      lcid == 0x0415 ||
      lcid == 0x041b) {
      usCodePage = 2;
    } else 
    if (lcid == 0x040d) {
      usCodePage = HEBREW_CP_Partial;
    } else 
    if (LOBYTE(lcid) == 0x01 || lcid == 0x0429) {
      usCodePage = ARABIC_CP_Partial;
    }
    else {
      // this is the default. (US english, 1252)
      usCodePage = 1;
    }


    // Get the partial base table
    pbPartialTable = g_prgbPartialBaseTbl[usCodePage][usSystem];

    // Generate the table.
    //
    // the first 128 characters is identity (except for 90 to 122)
    for (i=0; i < 128; rgchTbl[i]=i,  i++);

    // fix up the characters between 97 and 122 (i. A-Z maps to a-z)
    for (i=0; i < 26; rgchTbl[97+i] = 'A'+i, i++);

    // Get the top 128 characters from the partial table
    for (i=0;
   i < 128;
   rgchTbl[128+i] = pbPartialTable[i], i++);
    // end of FOR loop.

    // Now check if the is any exception for this lcid.
    i = 0;
    while ((g_rgLcidToExcepTbl[i][usSystem].m_lcid != 0) &&
     (g_rgLcidToExcepTbl[i][usSystem].m_lcid != lcid)) {
      i++;
    }

    // check if there is an exception table for this lcid
    if (g_rgLcidToExcepTbl[i][usSystem].m_lcid != 0) {
      // there is an exception table, hence we need to do some more fix up
      pbExcepTable = (BYTE *)g_rgLcidToExcepTbl[i][usSystem].pvExcepTbl;

      i = 0;
      while (*(pbExcepTable + 2*i) != 0) {
  rgchTbl[ (0x00ff & *(pbExcepTable + 2*i)) ] =
              (0x00ff & *(pbExcepTable + 2*i + 1));
  i++;
      } // while

    } // if

    // DONE !!!
    return;
}
#pragma code_seg()


